package applet.proxy;

import java.net.*;
import java.io.*;
import java.util.StringTokenizer;

import salsa.language.Actor;
import salsa.language.Message;
import salsa.language.RunTime;
import salsa.language.ServiceFactory;
import salsa.naming.UAL;
import salsa.messaging.ReceptionService;

import wwc.messaging.RMSPAcknowledgement;

import applet.theater.AddRequest;
import applet.theater.ResolveRequest;

public class ProxyReceptionService implements ReceptionService {
	private ObjectInputStream inputStream = null;
	private ObjectOutputStream outputStream = null;

	public ProxyReceptionService() {}

	public void processResolveRequest(ResolveRequest request) {
		// Send the resolved UAL back
		try {
			outputStream.writeObject( ServiceFactory.getNaming().resolve(request.uan) );
			outputStream.flush(); 
		} catch (IOException e) {
			System.err.println("Reception Service error: ");
			System.err.println("\tResolving Actor Name: " + request.uan);
			System.err.println("\tException: " + e.getMessage() );
		}
	}

	public void processAddRequest(AddRequest request) {
		ServiceFactory.getNaming().add(request.uan, request.ual);
	}

	public void processActor(Actor incomingState) {
		//add the state to the stateRegistry, so incoming actor references can
		//get to it
		ServiceFactory.getTheater().setEntry(incomingState.getUAN(), incomingState);
		incomingState.start();
		RunTime.receivedUniversalActor();

		// Send an RmspAcknowledgement.
		try {
			outputStream.writeObject(new RMSPAcknowledgement());
			outputStream.flush(); 
		} catch (IOException e) {
			System.err.println("Reception Service error: ");
			System.err.println("\tReceiving actor: " + incomingState.toString());
			System.err.println("\tCould not send RmspAcknowledgement, is the remote theater compliant?");
			System.err.println("\tException: " + e.getMessage() );
		}
	}

	public void processMessage(Message message) {
		UAL target = message.getTarget().getUAL();

		if (target.getLocation().equals( ServiceFactory.getTheater().getLocation() )) {
			//This message is intended for an actor on a applet that this proxy is servicing

			//parse the host and port from the UAL
			String targetHost = null;
			int targetPort = 0;

			StringTokenizer stringTokenizer = new StringTokenizer(target.getIdentifier(), "/");

			if (!stringTokenizer.nextToken().equals("applet")) {
				//The message was meant for an actor at this theater, but none should exist as it
				//is a proxy, so send a MessageTargetNotFound error message.

				//message.getSender()<-MessageTargetNotFound(message);
				Message NoTargetMessage = new Message(null, message.getSender(), "MessageTargetNotFound", null);
				Object[] arguments = { message };
				NoTargetMessage.setArguments( arguments );

				message.getSender().send(NoTargetMessage);
			} else {
				targetHost = stringTokenizer.nextToken();
				targetPort = new Integer( stringTokenizer.nextToken() ).intValue();

				try {
					Socket targetApplet = new Socket(targetHost, targetPort);
					ObjectOutputStream outputStream = new ObjectOutputStream( targetApplet.getOutputStream() );
					outputStream.writeObject( message );
					outputStream.close();
					targetApplet.close();
				} catch (Exception e) {
					System.err.println("ProxyReceptionService Error:");
					System.err.println("\tError: " + e);
					System.err.println("\tException with: " + targetHost + ":" + targetPort);
					System.err.println("\tBy Actors: " + message.toString());
					System.err.println("\tfor message: " + message.getDescriptor());
				}
			}
		} else {
			//This message is coming from an applet and intended for an actor on the normal WWC
			message.getTarget().send(message);
		}
	}

	public void process(Socket incoming) {
		// Read in the object.
		Object received = null;

		try {
			inputStream = new ObjectInputStream( incoming.getInputStream() );
			outputStream = new ObjectOutputStream( incoming.getOutputStream());
		} catch (IOException e) {
			System.out.println(e.getStackTrace().toString());
			System.err.println("IOException occured opening ObjectInputStream " + e);
		}

		while(true){
			try {								
				received = inputStream.readObject();
			} catch (ClassNotFoundException e) {
				System.err.println( "Could not load class for " + received.getClass().getName() );
				System.err.println( "  Is it in the theater CLASSPATH" );
				System.err.println( e.getMessage() );
				return;
			} catch (IOException e) {
				System.err.println( "IOException occured reading object " + e.getMessage() );
				return;
			}

			if (received instanceof Actor) {
				Actor incomingActor = (Actor)received;
				processActor(incomingActor);
			} else if (received instanceof Message) {  // just serializing data
				Message message = (Message)received;								
				processMessage(message);
			} else if (received instanceof AddRequest) {
				AddRequest request = (AddRequest)received;
				processAddRequest(request);
			} else if (received instanceof ResolveRequest) {
				ResolveRequest request = (ResolveRequest)received;
				processResolveRequest(request);
			}
		}
	}	
}

