/*
 * Created on Sep 17, 2003
 *
 */
package applet.theater;

import java.io.ObjectOutputStream;
import java.io.ObjectInputStream;
import java.io.IOException;
import java.net.Socket;
import java.net.UnknownHostException;
import java.util.Vector;
import java.util.Hashtable;

import salsa.language.*;
import salsa.language.exceptions.*;
import salsa.naming.*;
import salsa.messaging.*;

import wwc.messaging.RMSPAcknowledgement;

/**
 * @author elmagk
 */

/**
 * This class is the reference implementation of the Persistent MessagingService. The
 * service is implemented as a RMSP client.
 * 
 * This implementation uses the NamingService interface and the salsa.language.
 * ServiceFactory for UAN resolution.
 * 
 * @author Travis, elmagk
 */
public class AppletTransportService implements TransportService {
	private ObjectInputStream inputStream;
	private ObjectOutputStream outputStream;
	private Socket socket;

	private String proxyHost;
	private int proxyPort;

	public AppletTransportService() {}

	public void setProxyInformation(String host, int port) {
		proxyHost = host;
		proxyPort = port;
	}
	
	private void setInputStream() throws UnknownHostException, IOException {
		if (socket == null || socket.isClosed() || !socket.isConnected()) {
			//Create a new socket, either an exisiting socket has been disconnected or closed
			//or a new socket needs to be created since this is the first time we are contacting
	    		//this location

			try {
				socket = new Socket(proxyHost, proxyPort);
				socket.setKeepAlive(true);
			} catch (UnknownHostException e) {
				System.err.println("Unknown host: " + proxyHost + ":" + proxyPort);
				System.err.println("\terror: " + e);
			} catch (IOException e) {
				System.err.println("Error in creating a socket for an output stream: " + e);
				System.err.println("\twith target: " + proxyHost + ":" + proxyPort);
			}
			
			try {
				inputStream = new ObjectInputStream(socket.getInputStream());
			} catch (IOException e) {
				System.err.println("Error generating an ObjectOutputStream from socket: " + e);
				System.err.println("\twith target: " + proxyHost + ":" + proxyPort);
			}
		} else if (inputStream == null) {
			try {
				inputStream = new ObjectInputStream(socket.getInputStream());
			} catch (IOException e) {
				System.err.println("Error generating an ObjectOutputStream from socket: " + e);
				System.err.println("\twith target: " + proxyHost + ":" + proxyPort);
			}
		}
	}
	
	private void setOutputStream() throws UnknownHostException, IOException {
		if (socket == null || socket.isClosed() || !socket.isConnected()) {
			//Create a new socket, either an exisiting socket has been disconnected or closed
			//or a new socket needs to be created since this is the first time we are contacting
	    		//this location

			try {
				socket = new Socket(proxyHost, proxyPort);
				socket.setKeepAlive(true);
			} catch (UnknownHostException e) {
				System.err.println("Unknown host: " + proxyHost + ":" + proxyPort);
				System.err.println("\terror: " + e);
			} catch (IOException e) {
				System.err.println("Error in creating a socket for an output stream: " + e);
				System.err.println("\twith target: " + proxyHost + ":" + proxyPort);
			}
			
			try {
				outputStream = new ObjectOutputStream(socket.getOutputStream());
			} catch (IOException e) {
				System.err.println("Error generating an ObjectOutputStream from socket: " + e);
				System.err.println("\twith target: " + proxyHost + ":" + proxyPort);
			}
		} else if (outputStream == null) {
			try {
				outputStream = new ObjectOutputStream(socket.getOutputStream());
			} catch (IOException e) {
				System.err.println("Error generating an ObjectOutputStream from socket: " + e);
				System.err.println("\twith target: " + proxyHost + ":" + proxyPort);
			}
		}
	}

	public void migrate(Actor state, UAL target){
		try {
			setOutputStream();
			synchronized(outputStream){
				outputStream.writeObject( state );
				outputStream.flush();
			}
		} catch (IOException e) {
			throw new SalsaWrapperException("Error migrating actor: ", e);
		}

		// Wait for the Acknowledgement that the actor is situated on the remote theater.
		try {
			setInputStream();
			synchronized(inputStream){
				RMSPAcknowledgement ack = (RMSPAcknowledgement) inputStream.readObject();
			}
		} catch (Exception e) {
			throw new SalsaWrapperException( "Did not recieve RmspAcknowledgement. ", e);
		}
	}

	public void send(Message message, UAL target) {
		// Determine the destination of this message.
		if (target == null) {
			UAN uan = message.getTarget().getUAN();

			if (uan != null) {
				Object entry = ServiceFactory.getTheater().getEntry(uan);
				if ((entry instanceof UAL) || (entry == null)) {
					// Resolve the location.
					target = ServiceFactory.getNaming().resolve( uan );
					ServiceFactory.getTheater().setEntry(uan, target);
				} else {
					((Actor)entry).send(message);
					return;
				}
			} else {
				System.err.println("UAN and UAL of message are null, cannot send message");
			}
		}

		try {
			setOutputStream();
			synchronized(outputStream){
				outputStream.writeObject( message );
				outputStream.flush();
			}
		} catch (IOException e) {
			throw new SalsaWrapperException("Error sending message: ", e);
		}
	}

	public void send(Vector messages) {
		//used by the compiler to send a vector of messages generated by a message

		for (int i = 0; i < messages.size(); i++) {
			final Message current = (Message)messages.get(i);
			final ActorReference target = current.getTarget();

			try {
				if (current.getProperty() != null && current.getProperty().equals("delay")) {
					//This supports sending messages with the delay property
					Object[] parameters = current.getPropertyParameters();
					final long sleepTime = ((Integer)parameters[0]).intValue();

					Runnable delayThread = new Runnable() {
						public void run() {
							try {
								Thread.sleep(sleepTime);
							} catch (Exception e) {
								System.err.println("Error in delay thread for message: " + current.getName());
							}
							target.send(current);
						}
					};
					new Thread(delayThread, "Delay Thread").start();
				} else {
					target.send( current );
				}
			} catch (Exception e) {
				System.err.println("Transport service error: ");
				System.err.println("\tSending message: " + current.getDescriptor());
				if (current.getSender() != null) System.err.println("\tBy actor: " + current.getSender().toString());
				else System.err.println("\tBy actor: unknown");
				System.err.println("\tTo target: " + current.toString());
				System.err.println("\tException: " + e);
			}
		}
	}
}



