/*
 *	Modified Extensively for SALSA 0.6.0 new token semantics.
 *
 *	Author: Travis Desell
 *
 *	Generates the input and output tokens as generated by the salsa
 *	compiler's message sending code.
 */

package salsa.language;

import java.util.Vector;

public class Token implements java.io.Serializable {
	private String tokenName;

	/**
	 * If a token was specified a value at creation
	 * (through code like:  token x = 5), this method
	 * is used by the method to retrieve that value instead
	 * of waiting for a token.
	 */
	private Object	value = null;
	public	Object	getValue() 		{ return value; }

	/**
	 * The compiler specifies that this token is part of
	 * a join continuation by setJoinDirector.  Message then
	 * uses isJoinDirector to determine if this token is
	 * part of a join continuation or not.
	 */
	private boolean	isJoinDirector = false;
	public	boolean	isJoinDirector()	{ return isJoinDirector; }
	public	void	setJoinDirector() {
		if (isJoinDirector) {
			System.err.println("Token Creation Error:");
			System.err.println("\tAttempted to nest join blocks, currently this is not supported.");
			System.err.println("\tThis will cause errors with the current join block and it's return value.");
			System.err.println("\tWith token: " + tokenName);
		}

		isJoinDirector = true;
	}

	/**
	 * If this token is used as a token for a join director
	 * it needs to keep track of how many inputs that join
	 * director has.  This method is used by message to specify
	 * how many there are.
	 */
	private int joinInputs = 0;
	public int getJoinPosition() {
		joinInputs++;
		return (joinInputs - 1);
	}

	/**
	 * These methods create a join director and return a reference
	 * to that join director so tokens can be resolved to it.  The
	 * join director is passed a reference to this token, so it
	 * knows where to send its return value to.
	 */
	private ActorReference joinDirector = null;
	public void createJoinDirector() {
		joinDirector = (JoinDirector)new JoinDirector().construct(joinInputs, targetActors, targetMessages, targetPositions);
	}
	public ActorReference getJoinDirector() {
		return joinDirector;
	}


	/**
	 * These vectors are used to store the information about
	 * where to send tokens, and how they are resolved.
	 */
	private Vector targetActors = new Vector();
	private Vector targetMessages = new Vector();
	private Vector targetPositions = new Vector();

	/**
	 * This method is used by message to add an additional target
	 * to this token.
	 */
	public void addTarget(ActorReference actorReference, String messageId, int position) {
		targetActors.add(actorReference);
		targetMessages.add(messageId);
		targetPositions.add( new Integer(position) );
	}

	/**
	 * When a message has been processed and a value has been
	 * returned, this sends that value to each actor which has
	 * a message which it is being used in.
	 */
	public void resolve(ActorReference source, Object value) {
		for (int i = 0; i < targetActors.size(); i++) {
			ActorReference target = (ActorReference)targetActors.get(i);
			String messageId = (String)targetMessages.get(i);
			Integer position = (Integer)targetPositions.get(i);

			/**
			 * If the value is a large object, it isn't required to be sent
			 * if this token is being used for synchronization.  We can send
			 * null instead.
			 */
			Object[] arguments = new Object[3];
			arguments[0] = messageId;
			arguments[1] = position;
			if (position.intValue() == -3) arguments[2] = null;
			else arguments[2] = value;

			target.send( new Message(source, target, "resolveToken", arguments, null, null) );
		}
	}

	/**
	 *	These are the constructors for Token.  They specify the
	 *	tokens name, as well as retrieve a unique ID for this
	 * 	token.  If the token has a specified value it is set.
	 */
	public Token() {
		tokenName = "unnamed";
	}

	public Token(String tokenName) {
		this();
		this.tokenName = tokenName;
	}

	public Token(String tokenName, Object value) {
		this(tokenName);
		this.value = value;
	}
}
