
/**
 * $Id: UniversalActor.java,v 1.24 2004/04/22 13:18:08 deselt Exp $
 *
 * This is the main Actor class.
 */

package salsa.language;
import java.io.Serializable;
import java.io.IOException;
import java.io.ObjectOutputStream;
import java.io.ObjectInputStream;

import java.lang.reflect.Method;

import java.util.Hashtable;
import java.util.Vector;

import salsa.language.exceptions.CurrentContinuationException;

import salsa.messaging.TheaterService;
import salsa.messaging.TransportService;

import salsa.naming.NamingService;
import salsa.naming.UAL;
import salsa.naming.UAN;
import salsa.naming.MalformedUALException;
import salsa.naming.MalformedUANException;

import salsa.resources.OutputService;
import salsa.resources.ErrorService;
import salsa.resources.SystemService;


public class UniversalActor implements ActorReference, java.io.Serializable {
	private transient TransportService transportService = ServiceFactory.getTransport();
	private transient NamingService namingService = ServiceFactory.getNaming();

	private UAN uan = null;
	private UAL ual = null;
	/*
	 * These methods retrieve the UAN and UAL of the actor
	 */
	public UAN getUAN() { return uan; }
	public void setUAN(UAN uan) { this.uan = uan; }

	public UAL getUAL() { return ual; }
	public void setUAL(UAL ual) { this.ual = ual; }

	/*
	 * These new constructors have replaced getReferenceByName and getReferenceByLocation
	 * They prevent messy actor creation and syntactically make more sense.
	 * By default, the compiler uses the ActorReference(UAL) constructor, as all actorreferneces 
	 * begin as anonymous actors, and become non-anonymous when they are bound to a UAN
	 */
	public UniversalActor construct() { return this; }

	public void createRemotely(UAN uan, UAL targetUAL, String actorName) {
		RunTime.receivedUniversalActor();
		this.uan = uan;
		ual = namingService.generateUAL();

		/**
		 * First we create a placeholder at the current theater, and
		 * the actor at the remote theater.  After the actor at the
		 * remote theater has finished being created, we can remove
		 * the placeholder, and send the messages to the remote
		 * actor
		 */
		namingService.setEntry(uan, ual, new Placeholder(uan, ual));
		namingService.update(uan, ual);

		UniversalActor remoteSystem = (UniversalActor)UniversalActor.getReferenceByLocation( targetUAL.getLocation() + "salsa/System" );
		SystemService localSystem = ServiceFactory.getSystem();
		UniversalActor self = null;

		if (uan != null) {
			self = (UniversalActor)UniversalActor.getReferenceByName(uan);
			self.setUAL(ual);
		}
		else {
			System.err.println("Remote Actor Creation error:");
			System.err.println("\tImproper creation request for actor: " + toString());
			System.err.println("\tException: Cannot create an actor remotely if it does not have a UAN.");
			return;
		}
		//remoteTheater<-createActor(uan, ual, actorName) @
		//localTheater<-removePlaceholder(uan, ual);
		Token token1 = new Token();

		Object[] createActorArgs = { uan, targetUAL, actorName };
		Message createActorMsg = new Message(self, remoteSystem, "createActor", createActorArgs, null, token1);

		Object[] removePlaceholderArgs = { uan, ual };
		Message removePlaceholderMsg = new Message(self, localSystem, "removePlaceholder", removePlaceholderArgs, token1, null);

		remoteSystem.send(createActorMsg);
		localSystem.send(removePlaceholderMsg);
	}


	public UniversalActor()		{}
	public UniversalActor(UAN uan)	{
		//called by getReferenceByName
		this.uan = uan;
		this.ual = ServiceFactory.getNaming().get(uan);
	}
	public UniversalActor(UAL ual)	{ this.ual = ual; }

	/**
	 * The following methods provide a way to get a reference to an actor by name or by
	 * location.
	 */
	public static ActorReference getReferenceByName(String uan) throws MalformedUANException {
		return UniversalActor.getReferenceByName( new UAN(uan) );
	}
	public static ActorReference getReferenceByName(UAN uan) {
		return new UniversalActor(uan);
	}

	public static ActorReference getReferenceByLocation(String ual) throws MalformedUALException {
		return UniversalActor.getReferenceByLocation( new UAL(ual) );
	}
	public static ActorReference getReferenceByLocation(UAL ual) {
		return new UniversalActor(ual);
	}

	/* 
	 * The following cover local and remote message sending for Actors
	 */
	public void send(Message message) {
		Object target = namingService.getTarget(this);

		if (target instanceof Actor) {
			try {
				((Actor)target).putMessageInMailbox(message);
			} catch (Exception e) {
				System.err.println("Error putting message in mailbox: ");
				System.err.println("Exception: " + e);
				System.err.println("\t Message: " + message);
				e.printStackTrace();
			}
		} else if (target instanceof UAL) {
			this.ual = (UAL)target;
			transportService.send(message,this);
		} else {
			System.err.println("Message Sending Error:");
			System.err.println("\tActorReference: " + this.toString());
			System.err.println("\tMessage: " + message.toString());
			System.err.println("\tNamingService could not determine target for message.");
		}
	}

	/*
	 * Methods for printing out the identity of an actor
	 */
	public String toString() {
		String name = "";
		name +=  getClass().getName() + ": ";
		if (uan != null) name += uan.toString() + ", ";
		else name += "null, ";
		if (ual != null) name += ual.toString();
		else name += "null";
		return name;
	}

	public boolean equals(Object o) {
		if (o instanceof ActorReference) {
			ActorReference ref = (ActorReference)o;

			if (uan != null && ref.getUAN() != null) return uan.equals( ref.getUAN() );
			else if (ual != null && ref.getUAL() != null) return ual.equals( ref.getUAL() );
		} else if (o instanceof Actor) {
			Actor ref = (Actor)o;

			if (uan != null && ref.getUAN() != null) return uan.equals( ref.getUAN() );
			else if (ual != null && ref.getUAL() != null) return ual.equals( ref.getUAL() );
		}
		return false;
	}

	/*
	 * Methods for Actor serializability
	 */
	private void writeObject(java.io.ObjectOutputStream out) {
		//sending reference to local anonymous actor
		//need to keep the theater alive now since the reference has been
		//sent to the rest of the theater system.
		if (uan == null) RunTime.receivedUniversalActor();

		try {
			out.writeObject(uan);
			out.writeObject(ual);
			out.flush();
		} catch (IOException e) {
			System.err.println("Error during actor reference serialization: ");
			System.err.println("\tBy actor reference: " + toString());
			System.err.println("\tException: " + e);
		}
	}

	private void readObject(java.io.ObjectInputStream in) {
		try {
			uan = (UAN)in.readObject();
			ual = (UAL)in.readObject();
		} catch (Exception e) {
			System.err.println("Error during actor reference deserialization: ");
			System.err.println("\tBy actor reference: " + toString());
			System.err.println("\tException: " + e);
		}
		//get new instances of the services used by the Actor Reference
		transportService = ServiceFactory.getTransport();
		namingService = ServiceFactory.getNaming();
	}

/**Begin UniversalActor.State**/

public abstract class State extends Thread implements Actor, java.io.Serializable {
	private UAN uan;
	private UAL ual;
	public	UAN getUAN()		{ return uan; }
	public	UAL getUAL()		{ return ual; }
	public void setUAL(UAL ual)	{ this.ual = ual; }

	private transient NamingService namingService;
	private transient TheaterService theaterService;
	public	transient OutputService standardOutput;
	public	transient ErrorService standardError;

	public Vector __messages = new Vector();
	public Message currentMessage = null;

	String[] classNames = new String[0];
	public transient Hashtable methods = new Hashtable();


	public State(UAN uan, UAL ual) {
		namingService = ServiceFactory.getNaming();
		theaterService = ServiceFactory.getTheater();

		this.uan = uan;
		/**
		 * If no ual is specified, we need to generate one.
		 * if the specified ual is for the current theater,
		 * we can simply assign that ual to the actor. if not
		 * we need to migrate the actor to where it's supposed to be.
		 */
		if (uan == null && ual != null && !ual.getLocation().equals(theaterService.getLocation())) {
			System.err.println("Actor Creation Error:");
			System.err.println("\tAttempting to create a non-local actor without a UAN");
			System.err.println("\tActor: " + toString());
			return;
		}

		this.ual = ual;
		if (this.ual == null) this.ual = namingService.generateUAL();

		addClassName("salsa.language.UniversalActor$State");
		setName(getClass().getName());
	}

	public Method[] getMatches(String methodName) {
		return (Method[])methods.get(methodName);
	}

	public void addMethod(String methodName, Method method) {
		Method[] addedMethods = (Method[])methods.get(methodName);

		if (addedMethods == null) {
			addedMethods = new Method[1];
			addedMethods[0] = method;
			methods.put(methodName, addedMethods);
		} else {
			Method[] newAddedMethods = new Method[addedMethods.length + 1];
			newAddedMethods[0] = method;
			System.arraycopy(addedMethods, 0, newAddedMethods, 1, addedMethods.length);
			methods.put(methodName, newAddedMethods);
		}
	}

	public void addClassName(String className) {
		String[] tempNames = new String[classNames.length+1];
		tempNames[0] = className;
		System.arraycopy(classNames, 0, tempNames, 1, classNames.length);
		classNames = tempNames;
	}

	public void addMethodsForClasses() {
		methods = new Hashtable();

		for (int i = 0; i < classNames.length; i++) {
			try {
				Method[] declaredMethods = Class.forName(classNames[i]).getDeclaredMethods();
				for (int j = 0; j < declaredMethods.length; j++) {
					String name = declaredMethods[j].getName();
					addMethod( name, declaredMethods[j] );
				}
			} catch (Exception e) {
				System.err.println("Error generating methods for invocation, something is horribly wrong.");
				System.err.println("\tActor: " + toString() );
				System.err.println("\tError: " + e);
				e.printStackTrace();
			}
		}
	}

	public void messageTargetNotFound(Message message) {
		if (message.getTarget() instanceof UniversalActor) {
			UniversalActor target = (UniversalActor)message.getTarget();
			UAN uan = target.getUAN();

			if (uan != null) {
				UAL ual = target.getUAL();
				namingService.refreshReference(target);
				UAL newUAL = target.getUAL();

				if ( !ual.equals(newUAL) || (namingService.getTarget(target) instanceof Actor)) {
					target.send(message);
					return;
				}
			}
		}
		System.err.println("Message sending error:");
		System.err.println("\tMessage: " + message.toString());
		System.err.println("\tFrom: " + toString());
		System.err.println("\tTo: " + message.getTarget().toString());
		System.err.println("\tException: Could not find the target actor.");
		System.err.println("\t\tIt has no UAN and was not found at the theater of its UAL.");
	}

	/** Die kills the actor when its mailbox has been
	  * emptied
	  */
	private boolean dead = false;
	public	void 	die()		{ dead = false; }
	public	boolean	isDead()	{ return dead; }

	/** Destroy immediately kills the actor without processing any
	  * more messages
	  */
	private boolean destroyed = false;
	public	void	destroy()	{ destroyed = true; }
	public	boolean	isDestroyed()	{ return destroyed; }

	/**
	 * Mailbox is used to store messages with no unresolved tokens.
	 * pendingMessages stores all the messages with unresolved tokens.
	 * unresolvedTokens stores the tokens that made it to the actor beforee
	 * their target message.
	 */
	public Vector mailbox = new Vector();
	public Hashtable pendingMessages = new Hashtable();
	public Vector unresolvedTokens = new Vector();
	/**
	 * This class wraps the information from a resolveToken message,
	 * into an object so only one Vector is needed to store unresolved
	 * tokens.
	 */
	private class UnresolvedTokenWrapper implements java.io.Serializable {
		protected String	messageId;
		protected int		position;
		protected Object	value;
		
		public UnresolvedTokenWrapper(String messageId, int position, Object value) {
			this.messageId = messageId;
			this.position = position;
			this.value = value;
		}
	}

	/**
	 * resolveToken messages are sent when a message has completed processing, and
	 * returns a token that other messages are waiting for.
	 * This method updates the specified message with the given value, if the message
	 * has already been received, if not the information is placed in an
	 * UnresolvedTokenWrapper and put in a Vector of unresolved tokens.
	 */
	public synchronized void resolveToken(String messageId, int position, Object value) {
		/**
		 *	Try and resolve this token, if no matches are found
		 *	add it to the unresolvedHashtables because its
		 *	target message hasn't reached this actor yet.
		 */
		Message targetMessage = (Message)pendingMessages.get(messageId);

		if (targetMessage != null) {
			targetMessage.resolveToken(value, position);

			if (targetMessage.getRequiredTokens() == 0) {
				pendingMessages.remove(messageId);
				RunTime.finishedProcessingMessage();
			}
		} else {
			unresolvedTokens.add( new UnresolvedTokenWrapper(messageId, position, value) );			
		}
	}

	/**
	 *	This method puts a message in this actors mailbox,
	 *	if it is not waiting for any tokens. If it is, the
	 *	message goes in the Hashtable of pendingMessages.
	 */
	public synchronized void putMessageInMailbox(Message message) {
		RunTime.receivedMessage();

		/**
		 * If this message is waiting for any tokens, it cannot
		 * be put in the normal mailbox, because then it would be
		 * processed without those tokens being resolved. We put
		 * it in a seperate hashtable of pending messages instead.
		 */
		if (message.getRequiredTokens() > 0) {
			String messageId = message.getMessageId();
			pendingMessages.put(messageId, message);

			/**
			 * We need to check and see if any tokens that were meant for
			 * this message have already been receieved.  This requires
			 * going through the vector of unresolved tokens, and resolving
			 * the ones intended for this message.
			 */
			for (int i = 0; i < unresolvedTokens.size(); i++) {
				UnresolvedTokenWrapper unresolvedToken = (UnresolvedTokenWrapper)unresolvedTokens.get(i);

				if (unresolvedToken.messageId.equals(messageId)) {
					resolveToken(messageId, unresolvedToken.position, unresolvedToken.value);
					unresolvedTokens.remove(i);
					i--;
				}
			}
		} else {
			if (message.getMethodName().equals("construct") || (message.getProperty() != null && message.getProperty().equals("priority"))) {
			
				//This makes sure that the message doesn't preceeed messages with high priority
                                if(mailbox.size()!=0){
                                        int i = 0;
                                        while( i<mailbox.size() && (message.getProperty() != null && message.getProperty().equals("priority"))) i++;                              
                                        mailbox.insertElementAt(message, i);
                                                                                                                                                                                   
                                }
                                else
                                        mailbox.insertElementAt(message, 0);

			} else {
				mailbox.addElement(message);
			}

			if (message.getMethodName().equals("construct")) {
				start();
			} else {
				notify();
			}
		}
	}

	public synchronized Message getMessage() {
		if (mailbox.isEmpty()) {
			try {
				wait();      // The lock of Mailbox means it is empty.
			} catch (InterruptedException ie){
				System.err.println("Error from within salsa.language.Actor: " + toString());
				System.err.println("\tError getting a new message:");
				System.err.println("\t" + ie);
			}
		}
		Message nextMessage = (Message)mailbox.firstElement();
		mailbox.removeElementAt(0);

		return nextMessage;
	}

	public void run() {
		standardOutput = ServiceFactory.getOutput();
		standardError = ServiceFactory.getError();

		while ((!destroyed) && (!migrating) && (!dead || !mailbox.isEmpty()) ) {
			currentMessage = getMessage();
			process(currentMessage);
			RunTime.finishedProcessingMessage();
		}
	}

	public void sendGeneratedMessages() {
		/**
		 * When a message is processed, if it sends any messages they are placed in
		 * a vector called messages. After the messages are processed, this sends
		 * those messages to their targets.
		 */
		for (int i = 0; i < __messages.size(); i++) {
			final Message current = (Message)__messages.get(i);
			final ActorReference target = current.getTarget();
			/**
			 * If this message uses the delay property, this creates a thread to
			 * wait until that time has elapsed to send the message, if not the
			 * message is simply sent.
			 */
			if (current.getProperty() != null && current.getProperty().equals("delay")) {
				Object[] parameters = current.getPropertyParameters();
				final long sleepTime = ((Integer)parameters[0]).intValue();

				Runnable delayThread = new Runnable() {
					public void run() {
						try {
							Thread.sleep(sleepTime);
						} catch (Exception e) {
							System.err.println("Error Sending Message: ");
							System.err.println("\tError in delay thread caused by using the message delay property.");
							System.err.println("\tMessage: " + current.toString());
							System.err.println("\tException: " + e);
						}
						target.send(current);
					}
				};
				new Thread(delayThread, "Delay Thread").start();
			} else {
				target.send( current );
			}
		}
		__messages.clear();
	}

	/**
	 * This method moves the actor from one theater to another.
	 */
	private boolean migrating = false;
	public	boolean isMigrating() { return migrating; }

	public void migrate(String newUAL) throws MalformedUALException { 
		migrate(new UAL(newUAL));
	}

	public void migrate(UAL targetUAL) {
		//System.err.println("migrating to: " + targetUAL);

		// Update the name server.
		if (uan == null) {
			System.err.println("Actor migration error:");
			System.err.println("\tImproper migration request by: " + toString());
			System.err.println("\tBy message: " + currentMessage.toString());
			System.err.println("\tException: Cannot migrate an actor which does not have a UAN.");
			return;
		}

		/**
		 *	Stop the actor from processing more messages while we wait for it
		 *	to migrate.
		 */
		migrating = true;

		/**
		 * createPlaceholder() @ 
		 * remoteSystem<-addActor(this) @
		 * localSystem<-removePlaceholder() @
		 * currentContinuation;
		 */
		namingService.remove(uan, ual);
		namingService.setEntry(uan, ual, new Placeholder(uan, ual));

		UniversalActor remoteSystem = (UniversalActor)UniversalActor.getReferenceByLocation( targetUAL.getLocation() + "salsa/System" );
		SystemService localSystem = ServiceFactory.getSystem();
		UniversalActor self = (UniversalActor)UniversalActor.getReferenceByName(this.uan);

		Token token1 = new Token();

		Object[] addActorArgs = { this };
		Message addActorMsg = new Message(self, remoteSystem, "addActor", addActorArgs, null, token1);

		Object[] removePlaceholderArgs = { uan, ual };
		Message removePlaceholderMsg = new Message(self, localSystem, "removePlaceholder", removePlaceholderArgs, token1, currentMessage.getContinuationToken());

		ual = targetUAL;
		remoteSystem.send(addActorMsg);
		localSystem.send(removePlaceholderMsg);

		throw new CurrentContinuationException();
	}

	/**
	 * This prints a human readable version of the actor, with its uan and ual
	 * This will be equivalent to what a reference to this actor would return
	 * with toString().
	 */
	public String toString() {
		String name = "";
		name +=  getClass().getName() + ": ";
		if (uan != null) name += uan.toString() + ", ";
		else name += "null, ";
		if (ual != null) name += ual.toString();
		else name += "null";
		return name;
	}

	/**
	 * This returns true if two Actors are equal, it will also return true
	 * if an actor and a reference pointing to that actor are compared.
	 */
	public boolean equals(Object o) {
		if (o instanceof UniversalActor) {
			UniversalActor ref = (UniversalActor)o;

			if (uan != null && ref.getUAN() != null) return uan.equals(ref.getUAN());
			else if (ual != null && ref.getUAL() != null) return ual.equals(ref.getUAN());
		} if (o instanceof UniversalActor.State) {
                        UniversalActor.State ref = (UniversalActor.State)o;

			if (this.uan != null && ref.getUAN() != null) return this.uan.equals(ref.getUAN());
			else if (ual != null && ref.getUAL() != null) return ual.equals(ref.getUAN());
		}
		return false;
        }

	/** The following are used to serialize and deserialize an Actor
	 */
	private void writeObject(ObjectOutputStream out) throws IOException {
		out.defaultWriteObject();
		out.flush();
	}

	private void readObject(ObjectInputStream in) throws IOException, ClassNotFoundException {
		in.defaultReadObject();
		
		// Get service references from the ServiceFactory.
		namingService = ServiceFactory.getNaming();
		theaterService = ServiceFactory.getTheater();

		standardOutput = ServiceFactory.getOutput();
		standardError = ServiceFactory.getError();

		addMethodsForClasses();

		setName(getClass().getName());

		migrating = false;
		destroyed = false;
	}
}
/**END OF UniversalActor.State **/
}
