package wwc.naming;

import java.net.Socket;
import java.io.BufferedReader;
import java.io.PrintWriter;
import java.io.IOException;
import java.io.InputStreamReader;
import java.util.StringTokenizer;

import java.util.Hashtable;
import java.util.Vector;

import salsa.language.Actor;
import salsa.language.ActorReference;
import salsa.language.ServiceFactory;
import salsa.language.exceptions.SalsaException;
import salsa.language.exceptions.SalsaWrapperException;
import salsa.naming.NamingService;
import salsa.naming.UAL;
import salsa.naming.UAN;
import salsa.messaging.TheaterService;

/**
 * This class implements a UANP client. After the class is constructed, several
 * threads may pass through it to conduct UANP operations through the
 * NamingService interface in salsa.language.
 * 
 * @version $Id: WWCNamingService.java,v 1.6 2004/03/23 02:40:51 deselt Exp $
 * @author Abe Stephens -- stepha@rpi.edu
 * @author Travis Desell
 * @author Carlos Varela 
 */
public class WWCNamingService implements NamingService {

	public WWCNamingService() {}

        /**
         * These methods provide a way to generate a unique identifier
         * for every token used by a SALSA program and a default UAL for
         * new actors in the theater.
         */
	private int generatedTokens = 0;
	private int generatedLocations = 0;
	private TheaterService theater = ServiceFactory.getTheater();
	public synchronized String getUniqueMessageId() {
		return theater.getLocation() + "salsa/message" + (++generatedTokens) + "_" + System.currentTimeMillis();
	}
	public synchronized UAL generateUAL() {
		return new UAL(theater.getLocation() + "salsa/location" + (++generatedLocations) + "_" + System.currentTimeMillis());
	}


	/*
	 *	The UAL Registry contains UAL to UAN mappings in
	 *	the case of an actor with a universal name,
	 *	and UAL to actor mappings in the case of anonymous
	 *	actors.
	 *	Occasionally an actor will be renamed (UAL changed
	 *	without having a UAN), in this case the ualRegistry
	 *	has a mapping to a UAL.
	 */
	private Hashtable ualRegistry = new Hashtable();

	/*
	 *	The UAN Registry contains UAN to actor mappings for
	 *	local actors, and UAN to UAL Mappings for remote
	 *	actors.
	 */
	private Hashtable uanRegistry = new Hashtable();

	public synchronized Object getTarget(ActorReference actorReference) {
		UAN uan = actorReference.getUAN();
		Object retVal = null;

		if (actorReference.getUAN() == null && actorReference.getUAL() == null) {
			System.err.println("Message Sending Error:");
			System.err.println("\tActor Reference is invalid, has no UAN or UAL");
		}

		if (uan != null) {
			Object value = uanRegistry.get(uan.toString());

			if (value == null) {
				retVal = get(uan);
				uanRegistry.put(uan.toString(), retVal);
			} else retVal = value;
		} else {
			retVal = ualRegistry.get( actorReference.getUAL().toString() );
			if (retVal == null) retVal = actorReference.getUAL();
			else if (retVal instanceof UAN) {
				actorReference.setUAN((UAN)retVal);
				retVal = getTarget(actorReference);
			}
		}
		return retVal;
	}

	public synchronized void setEntry(UAN uan, UAL ual, Actor actor) {
		if (uan != null) {
			if (actor != null) uanRegistry.put(uan.toString(), actor);
			else uanRegistry.put(uan.toString(), ual);

			ualRegistry.put(ual.toString(), uan);
		} else {
			ualRegistry.put(ual.toString(), actor);
		}
	}

	public synchronized Actor remove(UAN uan, UAL ual) {
		Object value = null;

		if (ual != null) value = ualRegistry.remove(ual.toString());
		if (uan != null) value = uanRegistry.remove(uan.toString());

		if (value instanceof Actor) return (Actor)value;
		else return null;
	}

	public synchronized void refreshReference(ActorReference actorReference) {
		if (getTarget(actorReference) instanceof Actor) return;

		UAN uan = actorReference.getUAN();
		UAL newUAL = get(uan);

		if ( !(uanRegistry.get(uan) instanceof Actor) ) {
			actorReference.setUAL( newUAL );
			uanRegistry.put(uan.toString(), newUAL);
		}
	}

/*
 *	The following methods implement the basic functionality for
 *	accessing the Naming Server
*/
	/**
	 * This method provides UANP GET functionality.
	 * 
	 * @see salsa.language.NamingService#resolve(salsa.language.UAN)
	 */
	public UAL get(UAN name) {

		try {
			Socket s = new Socket(name.getHost(), name.getPort());
			
			BufferedReader in = new BufferedReader(new InputStreamReader(s.getInputStream()));
			PrintWriter out = new PrintWriter(s.getOutputStream(), true);
			out.println("GET " + name.getIdentifier() + " UANP/1.0\n" );

			String inputLine = in.readLine();
			if (inputLine == null) throw new SalsaException("UANP Server not responding.");

			StringTokenizer st = new StringTokenizer(inputLine);
			st.nextToken();
			
			String returnCode = new String(st.nextToken());
			if (!returnCode.equals("200"))
				throw new SalsaException( "UANP " + name + " Error: " + inputLine );

			return new UAL((new StringTokenizer(in.readLine())).nextToken());
		} catch (IOException e) {
			
			throw new SalsaWrapperException( "UANP GET Failed: " + name, e );
		} 
	}

	/**
	 * This method provides UANP PUT functionality.
	 * 
	 * @see salsa.language.NamingService#add(salsa.language.UAN, salsa.language.UAL)
	 */
	public void add(UAN name, UAL location) {
		
		try {
			Socket socket = new Socket(name.getHost(), name.getPort());
			
			BufferedReader in = new BufferedReader(new InputStreamReader(socket.getInputStream()));
			PrintWriter out = new PrintWriter(socket.getOutputStream(), true);
			
			out.println("PUT " + name.getIdentifier() + " " + location + " UAN/1.0\n");
			out.flush();
			String status = in.readLine();
		
			String response = in.readLine();

			in.close();
			out.close();
			socket.close();
			
		} catch (IOException e) {
			
			throw new SalsaWrapperException( "UANP PUT Failed: ", e );
		}		
	}

	/**
	 * This method is equivlent to add( name, location );
	 * 
	 * @see salsa.language.NamingService#update(salsa.language.UAN, salsa.language.UAL)
	 */
	public void update(UAN name, UAL location) {
		
		add( name, location );
	}

	/**
	 * This method is not implemented.
	 * 
	 * @see salsa.language.NamingService#delete(salsa.language.UAN, salsa.language.UAL)
	 */
	public void delete(UAN name, UAL location) {
		
		throw new SalsaException( "This naming service does not implement delete." );
	}

}
